'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _deepDiff = require('deep-diff');

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Returns true when two arrays of keys are equal
 * @param {[String]} actual - array of keys
 * @param {[String]} expected - array of expected keys
 * @returns {Boolean}
 */
function isPathsEqual(actual, expected) {
  return actual.length === expected.length && actual.every(function (k, i) {
    return k === expected[i];
  });
}

/**
 * Returns the value at the specified path in the target objecgt
 * @param {Object} target - target object
 * @param {[String]} path - array of nested keys
 * @returns {Mixed} value at the specified path
 */
function getSubjectAtPath(target, path) {
  return (typeof target === 'undefined' ? 'undefined' : _typeof(target)) !== 'object' ? null : path.reduce(function (value, key) {
    return typeof !value === 'undefined' ? value : value[key];
  }, target);
}

/**
 * Merges two `deep-diff` change objects as POJOs
 * @param {Object} a - first change to merge
 * @param {Object} b - second change to merge
 * @returns {Object} a new, merged change
 */
function mergeChanges(a, b) {
  // if an empty object is returned, there is no change
  var merged = {};

  // these changes would cancel out otherwise
  if (a.lhs !== b.rhs) {
    // arrays need their items merged
    if (a.kind === 'A' || b.kind === 'A') {
      var item = mergeChanges(a.kind === 'A' ? a.item : { kind: a.kind, lhs: a.lhs }, b.kind === 'A' ? b.item : { kind: b.kind, rhs: b.rhs });

      // items would cancel out otherwise
      if (item && item.kind) {
        // if `a` wasn't the array, `b` was
        merged = a.kind === 'A' ? _extends({}, a, { item: item }) : _extends({}, b, { item: item });
      }
      // if `b` was an edit and `a` was an addition, keep `a`
    } else if (a.kind !== 'D' && b.kind === 'E') {
      merged = _extends({}, a, { rhs: b.rhs });
      // if `a` was a delete and `b` isn't, it was an edit
    } else if (a.kind === 'D' && b.kind !== 'D') {
      merged = { kind: 'E', path: a.path, lhs: a.lhs, rhs: b.rhs };
      // if `a` wasn't a delete, but `b` was, keep `b`
    } else if (a.kind !== 'D' && b.kind === 'D') {
      merged = _extends({}, b, { lhs: a.lhs });
      // unhandled shouldn't be added to the diff
    } else {
      return;
    }
  }

  return merged;
}

/**
 * A diff accumulator is simply something that implements `.push()`
 *
 * This diff accumulator will flatten changes that match the `filter`
 * option. It will also automatically add those flattened changes to the
 * `prefilter` argument to prevent `deep-diff` from doing any further
 * analysis on the flattened object-property path.
 */

var DiffAccumulator = function () {

  /**
   * @constructor
   * @param {Function} flatten - matches changes that should be flattened
   * @param {Function} prefilter - matches changes that should not be analyzed
   */
  function DiffAccumulator() {
    var _this = this;

    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        flatten = _ref.flatten,
        prefilter = _ref.prefilter;

    _classCallCheck(this, DiffAccumulator);

    this.flattened = [];
    this.changes = [];
    this.lhs = null;
    this.rhs = null;

    this.prefilter = function (path, key) {
      if (_this.isFlattened([].concat(_toConsumableArray(path), [key]))) {
        return true;
      } else if (_this.opts.prefilter) {
        return _this.opts.prefilter(path, key);
      } else {
        return false;
      }
    };

    this.opts = { flatten: flatten, prefilter: prefilter };
  }

  /**
   * Wrap the `prefilter` option to also filter flattened changes
   * @param {[String]} path - object-property path
   * @param {String} key - current key to filter
   * @returns {Boolean} `true` to filter, `false` otherwise
   */


  _createClass(DiffAccumulator, [{
    key: 'diff',


    /**
     * Calculates the structural difference between two object, passing
     * this instance as the accumulator
     * @param {Object} lhs - left hand side assignment
     * @param {Object} rhs - right hand side aassignment
     * @returns {Array} the accumulated changes
     */
    value: function diff(lhs, rhs) {
      this.flattened = [];
      this.lhs = lhs;
      this.rhs = rhs;

      (0, _deepDiff.diff)(lhs, rhs, this.prefilter, this);
      return this.changes;
    }

    /**
     * Clears the current accumulated diffs
     */

  }, {
    key: 'clear',
    value: function clear() {
      this.changes = [];
    }

    /**
     * If the path of the change should be flattened, this method
     * returns the flattened path
     * @param {Object} change - `deep-diff` change object
     * @returns {[String]} the path of the flattened change
     */

  }, {
    key: 'getFlatPath',
    value: function getFlatPath(change) {
      var _this2 = this;

      var index = -1;

      if (this.opts.flatten) {
        index = change.path.findIndex(function (key, i, path) {
          return _this2.opts.flatten(path.slice(0, i), key);
        });
      }

      return index >= 0 ? change.path.slice(0, index + 1) : change.path;
    }

    /**
     * Returns `true` if the path or any parent path was flattened,
     * 'false' otherwise
     * @param {[String]} path - path to check
     * @returns {Boolean} `true` if any parent path was flattend
     */

  }, {
    key: 'isFlattened',
    value: function isFlattened(path) {
      return this.flattened.some(function (done) {
        return isPathsEqual(done, path);
      });
    }

    /**
     * Pushes a POJO change to the diff array, maybe merging previous changes
     * @param {Object} change - `deep-diff` change object
     */

  }, {
    key: 'addChange',
    value: function addChange(change) {
      var existingIndex = this.changes.findIndex(function (c) {
        // array paths should include the index when comparing across changes
        var pathA = change.kind === 'A' ? [].concat(_toConsumableArray(change.path), [change.index]) : change.path;
        var pathB = c.kind === 'A' ? [].concat(_toConsumableArray(c.path), [c.index]) : c.path;
        return isPathsEqual(pathA, pathB);
      });

      // ensure change is a POJO
      if (change.item) {
        change = _extends({}, change, { item: _extends({}, change.item) });
      } else {
        change = _extends({}, change);
      }

      // merge existing changes
      if (existingIndex > -1) {
        var merged = mergeChanges(this.changes[existingIndex], change);

        // remove the change if `merged` is empty
        if (merged && !merged.kind) {
          this.changes.splice(existingIndex, 1);
        } else if (merged) {
          this.changes.splice(existingIndex, 1, merged);
        }
      } else {
        this.changes.push(change);
      }
    }

    /**
     * Handles how changes are pushed to the diff. Will push flattened
     * changes when neccesary
     * @param {Object} change - `deep-diff` change object
     */

  }, {
    key: 'push',
    value: function push(change) {
      var flatPath = this.getFlatPath(change);
      if (this.isFlattened(flatPath)) return;

      // flatten non-array changes when the change path is deeper than
      // the potentially flattened path
      if (change.kind !== 'A' && !isPathsEqual(flatPath, change.path)) {
        var lhs = getSubjectAtPath(this.lhs, flatPath);
        var rhs = getSubjectAtPath(this.rhs, flatPath);
        this.addChange({ kind: 'E', path: flatPath, lhs: lhs, rhs: rhs });
        this.flattened.push(flatPath);
      } else {
        this.addChange(change);
      }
    }
  }]);

  return DiffAccumulator;
}();

exports.default = DiffAccumulator;