'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.default = createDeducer;

var _util = require('./util');

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

/**
 * Creates a function which is then able to deduce the state history
 * from a state containing a diff leaf
 * @param {Function} selector - given the arguments provided to the
 * resulting deducer, should return a desired part of the state
 * @param {Object} config - configuration object
 * @returns {Function} the deducer function
 */
function createDeducer(selector) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var _config$key = config.key,
      key = _config$key === undefined ? 'diff' : _config$key,
      _config$next = config.next,
      next = _config$next === undefined ? false : _config$next,
      _config$unique = config.unique,
      unique = _config$unique === undefined ? false : _config$unique,
      _config$index = config.index,
      index = _config$index === undefined ? false : _config$index,
      _config$range = config.range,
      range = _config$range === undefined ? false : _config$range,
      _config$limit = config.limit,
      limit = _config$limit === undefined ? false : _config$limit;


  if (index !== false && (range || limit) || range && limit) {
    console.warn('index, range, or limit should not be combined');
  }

  var makeChanges = next ? _util.applyChanges : _util.revertChanges;

  /**
   * Deduces the state history from a state containing a diff leaf.
   * Any additional arguments will be passed along to the `selector`
   * function
   * @param {Object} rawState - a state which contains a diff leaf
   * @returns {Array} an array containing the selected state history
   * @throws {Error} when the diff leaf cannot be found
   */
  var deducer = function deducer(rawState) {
    var history = rawState[key],
        state = _objectWithoutProperties(rawState, [key]);

    if (!history || !(history.prev && history.next)) {
      throw new Error('"' + key + '" is not a diff history object');
    }

    var diffs = next ? history.next : history.prev;

    // return from the cache if the diffs haven't changed
    if (deducer.cache.self) {
      var _deducer$cache$self = _slicedToArray(deducer.cache.self, 2),
          cached = _deducer$cache$self[0],
          result = _deducer$cache$self[1];

      if (diffs === cached) return result;
    }

    // normalize `range` and `limit` to lower and upper bounds
    var length = diffs.length;

    var _ref = range || [0, (limit || length) - 1],
        _ref2 = _slicedToArray(_ref, 2),
        lower = _ref2[0],
        upper = _ref2[1];
    // clamp the boundaries by the `index` or the diff length


    lower = index !== false ? index : Math.max(lower, 0);
    upper = index !== false ? index : Math.min(upper, length - 1);

    var deduced = [];

    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    for (var i = 0; i <= length; i++) {
      // make changes every time so previous/later states remain
      // accurate to the diff
      makeChanges(state, diffs[i]);

      // if we're outside the boundaries, don't do anything else
      if (i < lower || i > upper) continue;

      // initially load from the cache
      var cacheId = i - length;

      var _ref3 = deducer.cache[cacheId] || [],
          _ref4 = _slicedToArray(_ref3, 2),
          _cached = _ref4[0],
          _result = _ref4[1];

      // if the diff has changed, get the selected value and cache it


      if (!_cached || diffs[i] !== _cached) {
        _result = selector.apply(undefined, [state].concat(args));
        deducer.cache[cacheId] = [diffs[i], _result];
      }

      // skip results that follow each other
      if (unique) {
        var last = next ? deduced[deduced.length - 1] : deduced[0];
        if (_result === last) continue;
      }

      // push, unshift, or return the deduced result
      if (index === false) {
        if (next) {
          deduced.push(_result);
        } else {
          deduced.unshift(_result);
        }
      } else if (i === index) {
        return _result;
      }
    }

    // if the index never returned a result, we haven't deduced anything
    deduced = index === false ? deduced : null;
    // cache the deducers results
    deducer.cache.self = [diffs, deduced];
    return deduced;
  };

  deducer.cache = {};
  return deducer;
}