'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.default = diff;

var _cloneDeep = require('clone-deep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _accumulator = require('./accumulator');

var _accumulator2 = _interopRequireDefault(_accumulator);

var _actions = require('./actions');

var _util = require('./util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * Immutably prepends a new diff to the history's previous diffs
 * @param {Object} history - diff history
 * @param {Array} addition - new diff
 * @returns {Object} new diff history object
 */
function addToHistory(history, addition) {
  return addition.length === 0 ? history : _extends({}, history, {
    prev: [addition].concat(_toConsumableArray(history.prev)).slice(0, history.limit || history.prev.length + 1),
    next: []
  });
}

/**
 * Immutably moves previous diff history to the next diff history, optionally by
 * an offset
 * @param {Object} history - diff history
 * @param {Number} offset - amount of diffs to move
 * @returns {Object} new diff history object
 */
function jumpToPrevHistory(history) {
  var offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;

  offset = Math.min(history.prev.length, offset);

  return _extends({}, history, {
    prev: history.prev.slice(offset),
    next: [].concat(_toConsumableArray(history.prev.slice(0, offset).reverse()), _toConsumableArray(history.next)).slice(0, history.limit || history.next.length + offset)
  });
}

/**
 * Immutably moves the next diff distory to the previous diff history,
 * optionally by an offset
 * @param {Object} history - diff history
 * @param {Number} offset - amount of diffs to move
 * @returns {Object} new diff history object
 */
function jumpToNextHistory(history) {
  var offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 1;

  offset = Math.min(history.next.length, offset);

  return _extends({}, history, {
    prev: [].concat(_toConsumableArray(history.next.slice(0, offset).reverse()), _toConsumableArray(history.prev)).slice(0, history.limit || history.prev.length + offset),
    next: history.next.slice(offset)
  });
}

/**
 * Higher order reducer to track deep-diff states before and after each
 * action. Additonally will undo or redo the state when specific actions are
 * dispatched.
 * @param {Function} reducer - redux reducer
 * @param {Object} config - configuration object
 * @returns {Function} reduc reducer with a diff history leaf
 */
function diff(reducer) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var _config$key = config.key,
      key = _config$key === undefined ? 'diff' : _config$key,
      _config$limit = config.limit,
      limit = _config$limit === undefined ? 0 : _config$limit,
      _config$undoType = config.undoType,
      undoType = _config$undoType === undefined ? _actions.UNDO : _config$undoType,
      _config$redoType = config.redoType,
      redoType = _config$redoType === undefined ? _actions.REDO : _config$redoType,
      _config$jumpType = config.jumpType,
      jumpType = _config$jumpType === undefined ? _actions.JUMP : _config$jumpType,
      _config$skipAction = config.skipAction,
      skipAction = _config$skipAction === undefined ? function () {
    return false;
  } : _config$skipAction,
      _config$initialState = config.initialState,
      initialState = _config$initialState === undefined ? { prev: [], next: [] } : _config$initialState,
      _config$ignoreInit = config.ignoreInit,
      ignoreInit = _config$ignoreInit === undefined ? true : _config$ignoreInit,
      _config$flatten = config.flatten,
      flatten = _config$flatten === undefined ? function () {
    return false;
  } : _config$flatten,
      _config$prefilter = config.prefilter,
      prefilter = _config$prefilter === undefined ? function () {
    return false;
  } : _config$prefilter;

  // this will accumulate and merge diffs until `accum.clear()` is called

  var accum = new _accumulator2.default({ flatten: flatten, prefilter: prefilter });

  return function (rawState, action) {
    var _ref = rawState || {},
        history = _ref[key],
        state = _objectWithoutProperties(_ref, [key]);

    history = history || _extends({}, initialState, { limit: limit });

    var lhs = rawState && state;
    var rhs = reducer(lhs, action);
    var nextState = rhs || {};
    var changes = void 0,
        diffs = void 0;

    switch (action.type) {
      case undoType:
        nextState = (0, _util.revertChanges)((0, _cloneDeep2.default)(lhs), history.prev[0]);
        history = jumpToPrevHistory(history);
        break;

      case redoType:
        nextState = (0, _util.applyChanges)((0, _cloneDeep2.default)(lhs), history.next[0]);
        history = jumpToNextHistory(history);
        break;

      case jumpType:
        // apply a subset of previous diffs
        if (action.index < 0) {
          diffs = history.prev.slice(0, Math.abs(action.index));
          nextState = (0, _util.revertDiffs)((0, _cloneDeep2.default)(lhs), diffs);
          history = jumpToPrevHistory(history, Math.abs(action.index));

          // apply a subset of future diffs
        } else if (action.index > 0) {
          diffs = history.next.slice(0, Math.abs(action.index));
          nextState = (0, _util.applyDiffs)((0, _cloneDeep2.default)(lhs), diffs);
          history = jumpToNextHistory(history, action.index);
        }

        break;

      default:
        // when `rawState` is undefined, chances are this is the very first time
        // this reduce has been called. We ignore this initial call by default,
        // otherwise it will generate a diff for each leaf in the entire state.
        changes = rawState || !ignoreInit ? accum.diff(lhs, rhs) : [];

        if (!skipAction(action)) {
          history = addToHistory(history, changes);
          accum.clear();
        }
    }

    return _extends({}, nextState, _defineProperty({}, key, history));
  };
};